/*==========================================================*/
/*  RecordIO.cpp  -  demonstrate how to use Record I/O      */
/*                                                          */
/*      1. Creating and writing a (binary) record file      */
/*      2. File size                                        */
/*      3. Reading and rewriting a (binary) record file     */
/*                                                          */
/*  Author:   Bary W Pollack                                */
/*  Updated:  May 28, 2000                                  */
/*==========================================================*/

#include <iostream.h>
#include <fstream.h>
#include <iomanip.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

/*==========================================================*/
/*  Year in school and a few other constants                */
/*==========================================================*/

enum YEAR { FROSH=1, SOPH, JUNIOR, SENIOR, UNKNOWN };

char *sYear[] = { "", "Frosh", "Soph", "Junior", "Senior", "Unknown" };

const int NLTH = 13;
const int FLTH = 3;

/*==========================================================*/
/*  StudentRecord - this is the layout of the data file     */
/*==========================================================*/

struct StudentRecord
{
    long  nStudentNumber;
    char  sLastName [NLTH];
    char  sFirstName [NLTH];
    char  cInitial;
    YEAR  nYear;
    int   nEnterYear;
    bool  bActive;
};
const int SR_SIZE = sizeof StudentRecord;

/*==========================================================*/
/*  DisplayStudentRecord - display a single student's data  */
/*==========================================================*/

void DisplayStudentRecord (const int nSeq, const StudentRecord& sr)
{
    cout.setf (ios::right, ios::left);
    cout << setw (5) << nSeq << "   ";
    cout << setfill ('0') << setw (8) << sr.nStudentNumber << setfill (' ');
    cout.setf (ios::left);
    cout << setw (FLTH) << ' '
         << setw (NLTH) << sr.sLastName
         << setw (NLTH) << sr.sFirstName
         << setw (FLTH) << sr.cInitial
         << setw (NLTH) << sYear [sr.nYear]
         << setw (NLTH) << sr.nEnterYear
         << endl;
}

/*==========================================================*/
/*  First and Last names for students                       */
/*==========================================================*/

char *sFname [] = { "Sam", "John", "Mary", "Susan", "George", "Tony",
                    "Lou", "Jack", "Mickey", "Ken", "Lon",    "Hugh",
                    "Tina","Joy",  "Elton", "Joe",  "Rick",   "Lea" };

char *sLname [] = { "Smith", "Jones", "Wu", "Chen", "Madison", "Frank",
                    "Taylor", "Adams", "Newton", "Charles", "Sax", "Butler" };

const int NUM_FNAMES = (sizeof sFname / sizeof sFname[0]);
const int NUM_LNAMES = (sizeof sLname / sizeof sLname[0]);

/*==========================================================*/
/*  FYLE is the name of the data file that will be used     */
/*==========================================================*/

const char *FYLE = "Records.dat";

/*==========================================================*/
/*  Random - generate a random double on [0..1)             */
/*==========================================================*/

double Random (void)
{
    return double (rand ()) / RAND_MAX;
}

/*==========================================================*/
/*  IsPrime - predicate; true for a prime, false otherwise  */
/*==========================================================*/

bool IsPrime (const int n)
{
    bool bIsPrime = true;

    if (n > 3)
    {
        if (n % 2 == 0)
            bIsPrime = false;
        else
        {
            int nLimit = int (sqrt (n));
            for (int i = 3; i <= nLimit; i += 2)
                if (n % i == 0)
                {
                    bIsPrime = false;
                    break;
                }
        }
    }

    return bIsPrime;
}

/*==========================================================*/
/*  Generate - generate nNum student records containing     */
/*             more or less random data                     */
/*==========================================================*/

void Generate (const int nNum)
{
    ofstream outs;
    outs.open (FYLE, ios::out | ios::binary);
    if (! outs)
    {
        cerr << endl << "Couldn't create " << FYLE << endl;
        exit (1);
    }

    for (int i = 0; i < nNum; i++)
    {
        StudentRecord sr;
        sr.nStudentNumber = long (100000000L * Random ());
        strcpy (sr.sLastName,  sLname [int (NUM_LNAMES * Random ())]);
        strcpy (sr.sFirstName, sFname [int (NUM_FNAMES * Random ())]);
        sr.cInitial = char ('A' + (i % 26));
        sr.nYear = YEAR (1 + int (5 * Random ()));
        sr.nEnterYear = int (1991 + 10 * Random ());
        sr.bActive = true;
        outs.write ((char *) &sr, sizeof (sr));
    }
    outs.close ();
}

/*==========================================================*/
/*  FileSize - returns the (long) length of the file        */
/*             measured in bytes                            */
/*==========================================================*/

long FileSize (const char *sFileName
{
    long lFileSize = -1;
    ifstream ins;
    ins.open (sFileName, ios::in | ios::nocreate | ios::binary);
    if (! ins)
    {
        cerr << endl << "Couldn't open " << sFileName << endl;
        exit (1);
    }
    ins.seekg (0, ios::beg);
    lFileSize = ins.tellg ();
    ins.seekg (0, ios::end);
    lFileSize = ins.tellg () - lFileSize;
    ins.close ();

    return lFileSize;
}

/*==========================================================*/
/*  DisplayActive - display the entire file sequentially,   */
/*                  from beginning to end                   */
/*==========================================================*/

void DisplayActive (void)
{
    ifstream ins;
    ins.open (FYLE, ios::in | ios::nocreate | ios::binary);
    if (! ins)
    {
        cerr << endl << "Couldn't read " << FYLE << endl;
        exit (1);
    }

    cout << "Number StudentID   LastName     FirstName  Init Year    EnterYear" << endl
         << "====== =========   ============ ========== ==== ====    =========" << endl
         << endl;

    int n = 0;
    int nActive = 0;
    StudentRecord sr;

    while (ins.read ((char *) &sr, sizeof (sr)), !ins.fail ())
    {
        ++n;
        if (sr.bActive)
        {
            ++nActive;
            DisplayStudentRecord (n, sr);
        }
    }

    cout << endl 
         << n << " student records     " << nActive << " active students"
         << "     File size is " << FileSize (FYLE) << " bytes" << endl
         << endl;

    ins.close ();
}

/*==========================================================*/
/*  DeactivateStudents - deactivate selected students       */
/*                    we'll select records with prime       */
/*                    indices in descending order down to 1 */
/*==========================================================*/
/*  PositionFile - positions an open file so that you may   */
/*                 read the desired record                  */
/*==========================================================*/

void PositionFile (fstream& ins, const int n)
{
    unsigned uLocation = (n - 1) * SR_SIZE;
    ins.seekg (uLocation, ios::beg);
}

/*==========================================================*/

void DeactivateStudents (void)
{
    fstream inouts;
    inouts.open (FYLE, ios::in | ios::out | ios::nocreate | ios::binary);
    if (! inouts)
    {
        cerr << endl << "Couldn't read/write " << FYLE << endl;
        exit (1);
    }

    cout << "Number StudentID   LastName     FirstName  Init Year    EnterYear" << endl
         << "====== =========   ============ ========== ==== ====    =========" << endl
         << endl;

    int n = 0;
    StudentRecord sr;
    int nNumRecords = FileSize (FYLE) / SR_SIZE;

    for (int i = nNumRecords; i >= 1; i--)
    {
        if (IsPrime (i))
        {
            PositionFile (inouts, i);
            inouts.read ((char *) &sr, sizeof (sr));
            DisplayStudentRecord (i, sr);
            sr.bActive = false;
            PositionFile (inouts, i);
            inouts.write ((char *) &sr, sizeof (sr));
            ++n;
        }
    }

    cout << endl 
         << n << " student records were deactivated" << endl
         << endl;

    inouts.close ();
}

/*==========================================================*/
/*  main program - drive the demonstration                  */
/*==========================================================*/

int main (void)
{
    cout << endl << "Demonstrate Record I/O" << endl << endl;

    srand (42);         // seed the random number generator

    Generate (50);      // generate the (random) student data

    DisplayActive ();      // display all the (active) records

    DeactivateStudents (); // deactivate some students

    DisplayActive ();   // display all the (active) records

    cout << endl;
    return 0;
}

/*==========================================================*/

/*  .....  Output from the program:


Demonstrate Record I/O


Number StudentID   LastName     FirstName  Init Year    EnterYear
====== =========   ============ ========== ==== ====    =========

    1   00534073   Smith        Ken          A  Unknown      1995         
    2   39304788   Wu           George       B  Junior       1991         
    3   98907437   Sax          Jack         C  Soph         1991         
    4   70879238   Adams        Jack         D  Junior       1997         
    5   27182836   Charles      Rick         E  Senior       1993         
    6   21485030   Wu           Susan        F  Frosh        1992         
    7   92110965   Jones        Joe          G  Junior       1993         
    8   95056001   Madison      Rick         H  Soph         1996         
    9   83831293   Sax          Joe          I  Junior       1992         
   10   22275460   Jones        Lea          J  Senior       1995         
   11   19934690   Adams        Ken          K  Frosh        1992         
   12   90646076   Charles      Ken          L  Soph         1999         
   13   11523789   Wu           Joy          M  Unknown      1993         
   14   07281716   Jones        Lon          N  Senior       1991         
   15   53993346   Butler       Ken          O  Senior       1992         
   16   72771385   Charles      Lon          P  Frosh        1991         
   17   29767754   Wu           John         Q  Frosh        1998         
   18   93868831   Sax          George       R  Frosh        1998         
   19   02581865   Charles      Ken          S  Frosh        1996         
   20   13776055   Sax          Elton        T  Unknown      1992         
   21   95620593   Taylor       Joe          U  Junior       1996         
   22   53749198   Smith        Ken          V  Unknown      1992         
   23   91363261   Charles      Sam          W  Unknown      1999         
   24   01416058   Newton       Jack         X  Frosh        1997         
   25   81722464   Taylor       Rick         Y  Unknown      1996         
   26   11883907   Jones        Joy          Z  Junior       2000         
   27   45793023   Butler       Mary         A  Senior       1997         
   28   69850764   Jones        Tony         B  Unknown      2000         
   29   51411481   Wu           Sam          C  Senior       1998         
   30   57286294   Newton       Susan        D  Senior       2000         
   31   78460036   Charles      Lon          E  Soph         1991         
   32   44148075   Smith        Jack         F  Unknown      1997         
   33   59038056   Chen         Sam          G  Senior       1999         
   34   31354716   Chen         Mary         H  Soph         1999         
   35   34369945   Chen         John         I  Soph         1999         
   36   88894314   Smith        Mary         J  Unknown      1999         
   37   53578295   Frank        Lon          K  Senior       1991         
   38   09381389   Charles      Hugh         L  Frosh        1992         
   39   97454756   Madison      George       M  Unknown      1992         
   40   01489303   Smith        George       N  Unknown      1998         
   41   13705862   Smith        Jack         O  Unknown      1998         
   42   40861232   Frank        Tony         P  Unknown      1995         
   43   11230811   Wu           Mickey       Q  Frosh        1991         
   44   40849024   Sax          Jack         R  Senior       2000         
   45   92443617   Wu           Lon          S  Frosh        1992         
   46   64180425   Adams        Elton        T  Soph         2000         
   47   16949980   Chen         Sam          U  Junior       1993         
   48   32029175   Jones        John         V  Unknown      1993         
   49   99481185   Madison      Tony         W  Unknown      1995         
   50   61375774   Chen         Tony         X  Unknown      1993         

50 student records     50 active students     File size is 2200 bytes

Number StudentID   LastName     FirstName  Init Year    EnterYear
====== =========   ============ ========== ==== ====    =========

   47   16949980   Chen         Sam          U  Junior       1993         
   43   11230811   Wu           Mickey       Q  Frosh        1991         
   41   13705862   Smith        Jack         O  Unknown      1998         
   37   53578295   Frank        Lon          K  Senior       1991         
   31   78460036   Charles      Lon          E  Soph         1991         
   29   51411481   Wu           Sam          C  Senior       1998         
   23   91363261   Charles      Sam          W  Unknown      1999         
   19   02581865   Charles      Ken          S  Frosh        1996         
   17   29767754   Wu           John         Q  Frosh        1998         
   13   11523789   Wu           Joy          M  Unknown      1993         
   11   19934690   Adams        Ken          K  Frosh        1992         
    7   92110965   Jones        Joe          G  Junior       1993         
    5   27182836   Charles      Rick         E  Senior       1993         
    3   98907437   Sax          Jack         C  Soph         1991         
    2   39304788   Wu           George       B  Junior       1991         
    1   00534073   Smith        Ken          A  Unknown      1995         

16 student records were deactivated

Number StudentID   LastName     FirstName  Init Year    EnterYear
====== =========   ============ ========== ==== ====    =========

    4   70879238   Adams        Jack         D  Junior       1997         
    6   21485030   Wu           Susan        F  Frosh        1992         
    8   95056001   Madison      Rick         H  Soph         1996         
    9   83831293   Sax          Joe          I  Junior       1992         
   10   22275460   Jones        Lea          J  Senior       1995         
   12   90646076   Charles      Ken          L  Soph         1999         
   14   07281716   Jones        Lon          N  Senior       1991         
   15   53993346   Butler       Ken          O  Senior       1992         
   16   72771385   Charles      Lon          P  Frosh        1991         
   18   93868831   Sax          George       R  Frosh        1998         
   20   13776055   Sax          Elton        T  Unknown      1992         
   21   95620593   Taylor       Joe          U  Junior       1996         
   22   53749198   Smith        Ken          V  Unknown      1992         
   24   01416058   Newton       Jack         X  Frosh        1997         
   25   81722464   Taylor       Rick         Y  Unknown      1996         
   26   11883907   Jones        Joy          Z  Junior       2000         
   27   45793023   Butler       Mary         A  Senior       1997         
   28   69850764   Jones        Tony         B  Unknown      2000         
   30   57286294   Newton       Susan        D  Senior       2000         
   32   44148075   Smith        Jack         F  Unknown      1997         
   33   59038056   Chen         Sam          G  Senior       1999         
   34   31354716   Chen         Mary         H  Soph         1999         
   35   34369945   Chen         John         I  Soph         1999         
   36   88894314   Smith        Mary         J  Unknown      1999         
   38   09381389   Charles      Hugh         L  Frosh        1992         
   39   97454756   Madison      George       M  Unknown      1992         
   40   01489303   Smith        George       N  Unknown      1998         
   42   40861232   Frank        Tony         P  Unknown      1995         
   44   40849024   Sax          Jack         R  Senior       2000         
   45   92443617   Wu           Lon          S  Frosh        1992         
   46   64180425   Adams        Elton        T  Soph         2000         
   48   32029175   Jones        John         V  Unknown      1993         
   49   99481185   Madison      Tony         W  Unknown      1995         
   50   61375774   Chen         Tony         X  Unknown      1993         

50 student records     34 active students     File size is 2200 bytes


..... End of Output ..... */

